import serial as serial
import numpy as np

#destinations
DEST_CMD = 0
DEST_REG_4dt0 = 2**5
DEST_REG_9dt5 = 2**6
DEST_REG_14dt10 = 2**6+2**5
DEST_REG_19dt15 = 2**7
DEST_REG_24dt20 = 2**7+2**5
DEST_REG_29dt25 = 2**7+2**6
DEST_REG_31dt30 = 2**7+2**6+2**5

CMD_ADDR_MV = 0
CMD_DPHI_WR = 2**0
CMD_NMEAS_WR = 2**1
CMD_NRINGUP_WR = 2**1+2**0
CMD_NSWEEP_WR = 2**2
CMD_DDPHI_WR = 2**2+2**0
CMD_NFREQS_MV = 2**2+2**1+2**0
CMD_START = 2**3
CMD_READBACK = 2**3+2**0
CMD_RESET = 2**3+2**1

def vectToScalar(data):
    cumsum = 0
    for i in range(8):
        cumsum = cumsum*256 + data[7-i]
    return cumsum



class Stitch:
    def __init__(self, serialPort, alwaysOpen=False, baudRate=115200):
        self.serialPort = serialPort
        self.alwaysOpen = alwaysOpen
        self.baudRate = baudRate
        self.numFreqs = 1023
        if self.alwaysOpen:
            self.serialHandle = serial.Serial(self.serialPort, self.baudRate, bytesize=8, stopbits=1, timeout=10)

    #public interface starts here
    def setFrequencies(self, dphis):
        for current_address in range(len(dphis)):
            self.setAddress(current_address)
            self.setRegister(dphis[current_address])
            self.dphiWrite()

    def setMeasureDurations(self, times):
        for current_address in range(len(times)):
            self.setAddress(current_address)
            self.setRegister(times[current_address])
            self.nmeasWrite()
    
    def setRingupDurations(self, times):
        for current_address in range(len(times)):
            self.setAddress(current_address)
            self.setRegister(times[current_address])
            self.nringupWrite()
    
    def setSweepDurations(self, times):
        for current_address in range(len(times)):
            self.setAddress(current_address)
            self.setRegister(times[current_address])
            self.nsweepWrite()
    
    def setSweepFrequencyDeltas(self, ddphis):
        for current_address in range(len(ddphis)):
            self.setAddress(current_address)
            self.setRegister(ddphis[current_address])
            self.ddphiWrite()
    
    def setNumFrequencies(self, numFreqs):
        self.setRegister(numFreqs)
        self.nfreqsTransfer()
        self.numFreqs = numFreqs

    def startExperiment(self):
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_START
        self._sendArray(arr)

    def getData(self):
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_READBACK
        self._sendArray(arr)
        #now read until timeout
        if not self.alwaysOpen:
            self.serialHandle = serial.Serial(self.serialPort, self.baudRate, bytesize=8, stopbits=1, timeout=10)
        data = self.serialHandle.read(20*self.numFreqs)
        if not self.alwaysOpen:
            self.close()
        print(f"Received {int(len(data)/20)} frames")
        #print(data)
        cosineData = []
        sineData = []
        numberData = []
        for i in range(self.numFreqs):
            cosineData.append([data[i*20:i*20+8]])
            sineData.append([data[i*20+8:i*20+16]])
            numberData.append([data[i*20+16:i*20+20]])
        #for i in range(self.numFreqs):
            #cosData = int.from_bytes(data[i*20:i*20+8], byteorder='little')
            #sinData = int.from_bytes(data[i*20+8:i*20+16], byteorder='little')
            #numData = int.from_bytes(data[i*20+16:i*20+20], byteorder='little')
            #cosineData.append(float(cosData)/float(numData))
            #sineData.append(float(sinData)/float(numData))
            #numberData.append(numData)
        return numberData, cosineData, sineData



    #no need to call anything down here
    def setAddress(self, address):
        self.setRegister(address)
        #transfer register to address
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_ADDR_MV
        self._sendArray(arr)
    
    def setRegister(self, value):
        arr = np.zeros(7,dtype='uint8')
        #set 4:0 bits of register
        arr[0] = DEST_REG_4dt0 + (value&31)
        #set 9:5 bits of register
        value = value >> 5
        arr[1] = DEST_REG_9dt5 + (value&31)
        #set 14:10 bits of register
        value = value >> 5
        arr[2] = DEST_REG_14dt10  + (value&31)
        #set 19:15 bits of register
        value = value >> 5
        arr[3] = DEST_REG_19dt15  + (value&31)
        #set 24:20 bits of register
        value = value >> 5
        arr[4] = DEST_REG_24dt20  + (value&31)
        #set 29:25 bits of register
        value = value >> 5
        arr[5] = DEST_REG_29dt25  + (value&31)
        #set 31:30 bits of register
        value = value >> 5
        arr[6] = DEST_REG_31dt30  + (value&31)
        #checkValue = 0
        #for i,entry in enumerate(arr):
        #    checkValue += (int(entry)&31)*2**(i*5)
        #    print(f"Entry {i}: {entry} -> {entry&31} -> {int(entry)&31}")
        #print(arr)
        #print(int(checkValue))
        self._sendArray(arr)

    def dphiWrite(self):
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_DPHI_WR
        self._sendArray(arr)

    def ddphiWrite(self):
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_DDPHI_WR
        self._sendArray(arr)

    def nmeasWrite(self):
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_NMEAS_WR
        self._sendArray(arr)

    def nringupWrite(self):
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_NRINGUP_WR
        self._sendArray(arr)

    def nsweepWrite(self):
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_NSWEEP_WR
        self._sendArray(arr)

    def nfreqsTransfer(self):
        arr = np.zeros(1,dtype='uint8')
        arr[0] = CMD_NFREQS_MV
        self._sendArray(arr)
    
    def _sendArray(self, transferBuffer):
        if not self.alwaysOpen:
            self.serialHandle = serial.Serial(self.serialPort, self.baudRate, bytesize=8, stopbits=1, timeout=1)
            self.serialHandle.reset_input_buffer()
        self.serialHandle.write(transferBuffer.astype('ubyte'))
        if not self.alwaysOpen:
            self.close()
    
    def close(self):
        self.serialHandle.close()
